#include "JVizTimeCommandObj.h"
#include "JInteger.h"
#include "JIntegerData.h"
#include "JStringData.h"
#include "JIntegerProperty.h"
#include "JIntegerListProperty.h"
#include "JStringProperty.h"
#include "JLinkObj.h"
#include "JBytesData.h"


char* theJVizTimeCommandObj = JVizTimeCommandObj().Register();


void JVizTimeCommandObj::writeContent(JOutputStream& os) {
  JAddObj::writeContent(os);
  putObject(os, "command", command);
}

void JVizTimeCommandObj::readContent(JDictionary& dict) {
  JAddObj::readContent(dict);
  JObject* obj_command = getObject( dict, "command");
  if ( obj_command ) command = *(JString*)obj_command;
}

const char* JVizTimeCommandObj::className() const { return "JVizTimeCommandObj";}


JObject* JVizTimeCommandObj::clone() const { return new JVizTimeCommandObj(*this);}


// class constructor
JVizTimeCommandObj::JVizTimeCommandObj() {
  igm = 1;
  ogm = 1;
  tick = 0;
}


void JVizTimeCommandObj::draw(JGraphics g, int x, int y, int w, int h) {
  JRect rect(x, y, w, h);
  if (command.length() > 0 ) {
	  drawText(g, command, rect );
  } else {
	  drawText(g, JString("VizTime"), rect);
  }
}


boolean JVizTimeCommandObj::inputAllowed(int n)
{ return !inputSet(n).last();}
JString JVizTimeCommandObj::inputTag(int n) {
  char* tag[] = {
    "tick" };
  return( tag[n] );
}


int JVizTimeCommandObj::outputType( int n ) {
  if ( n == 0 ) return JStringData::id;
  return JIntegerData::id;
}
JString JVizTimeCommandObj::outputTag(int n) {
  char* tag[] = {
    "result" };
  return( tag[n] );
}


JFRect JVizTimeCommandObj::inputArea(int n) {
  return topArea( n, 0, 1, 0.25 );
}


JFRect JVizTimeCommandObj::outputArea(int n) {
  return rightArea(n, 0, 1);
}


int JVizTimeCommandObj::inputFace(int n) {
  return TOP;
}


int JVizTimeCommandObj::outputFace(int n) {
  return RIGHT;
}


JFPoint JVizTimeCommandObj::inputPoint(int n, JLinkObj& link) {
  return topPoint(n, link, 0, 1 );
}


JFPoint JVizTimeCommandObj::outputPoint(int n, JLinkObj& link) {
  return rightPoint(n, link, 0, 1);
}


void JVizTimeCommandObj::access(int n, JLinkObj& link, const JDataType& data) {
  if ( n == 0 ) {
    JSTRING(data) = result;
  }
}


void JVizTimeCommandObj::engine(int n, JLinkObj& link) {
  switch (n) {
  case TICK:
    link.access(JIntegerData(tick));
    result = JString( command );
	if ( tick ) {
		broadcast( 0 );
	}
    break;
  default:
    break;
  }
}


JArray JVizTimeCommandObj::getProperties() {
  JArray properties;
  properties.append(JStringProperty("command", command));
  return properties;
}


boolean JVizTimeCommandObj::updateProperty(JProperty& prop) {
  if (prop.getName() == JString("command")) {
    command = ((JStringProperty*)&prop)->value;
    return true;
  }
  return false;
}


